// Cache.cpp: implementation of the Cache class.
//
//////////////////////////////////////////////////////////////////////

#include "Cache.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Cache::Cache(int iSize)
{
  m_iSize = iSize;
  //m_iWarmup = iWarmup + m_iSize;
  m_bWarmuping = TRUE;
  m_vecStats.resize(End,0);
  strcpy(m_sName, "Cache");
  m_iAlg = -1;

  m_vecPLsPgCache.push_back(NULL); // no page has reference count 0
  m_vecPLsPgCache.push_back(new ListPage);
  m_iPageNum=0;
  m_iEvtNum = 0;
}

Cache::~Cache()
{
  int i;
  for (i=1; i<m_vecPLsPgCache.size(); i++)
    delete m_vecPLsPgCache[i];
}

//////////////////////////////////////////////////////////////////////
// Public methods
//////////////////////////////////////////////////////////////////////

// print out statistics data
void Cache::print(ofstream &of)
{
  int i;
  char s[10];

  sprintf(s,"%4s%4d ", m_sName, m_iSize);
  of << s;
  for (i=0; i<End; i++)
  {
    sprintf(s,"%6d ",m_vecStats[i]);
    of << s;
  }
  of << endl;
}

// process this event and return whether this event should be
//  propagated
BOOL Cache::propagate(PEvent pEvent)
{
  ItListHashNode cur;
  Page page(pEvent);
  BOOL ret = TRUE;

  m_iEvtNum++;
  /* for debug
  if (m_iEvtNum % 100 ==0)
  {
    if (m_iSize==1000) // only for server
    {
      static int pmiss=0;
      static int pread=0;
      cout << m_iEvtNum << "\t" << 
        (m_vecStats[RdMiss]-pmiss)*1.0/(m_iEvtNum-pread) << endl;
      pmiss = m_vecStats[RdMiss];
      pread = m_vecStats[Read];
    }
  }
  //*/
  LogStats(pEvent);

  if (pEvent->rw==ERASE)
    erase(&page);
  else if ((cur=m_hash.Search(&page))!=NULL) // this event is already in cache
  {
    m_itCur = cur->it;
    LogHit(pEvent);
    if (pEvent->rw==READ)
      ret = FALSE; // do not propagate for read hit

    assert(page == *(cur->it));
    updatePage(cur);
  }
  else // this page is not in the cache
  {
    LogMiss(pEvent);
    add(&page); // add this event to cache
  }
  //check(); // for debug
  return ret;
}

//////////////////////////////////////////////////////////////////////
// Protected methods
//////////////////////////////////////////////////////////////////////

// add pPage into the cache. evict a page if the page is full
void Cache::add(PPage pPage)
{
  ItListPage it;

  if (m_iSize == 0) // don't add for empty cache
    return; 

  if (m_iPageNum==m_iSize) // cache is full, evict one page
    evict();
  else
    m_iPageNum++;

  // add the page to the head
  m_vecPLsPgCache[1]->push_front(*pPage);
  m_hash.Add(m_vecPLsPgCache[1]->begin());
}

// check the whole cache and delete all blocks that have
// the same (dev, inode) pair with pPage
void Cache::erase(PPage pPage)
{
  ItListPage it, itprev;
  DWORD ref, size;

  size = m_vecPLsPgCache.size();
  m_iPageNum=0;
  for (ref=1; ref < size; ref++)
  {
    eraseFile(m_vecPLsPgCache[ref], pPage);
    m_iPageNum += m_vecPLsPgCache[ref]->size();
  }
}

// check one list and delete all blocks that have
// the same (dev, inode) pair with pPage
void Cache::eraseFile(PListPage pListPages, PPage pPage)
{
  ItListPage it, itprev;

  it=pListPages->begin();
  while (it!=pListPages->end())
  {
    itprev=it;
    it++;
    if ((itprev->dev ==pPage->dev) &&
        (itprev->inode==pPage->inode))
    {
      m_hash.Erase(&(*itprev));
      pListPages->erase(itprev);
    }
  }
}

int Cache::GetDepth(ItListPage itPage) // get the stack depth
{
  int ref, depth, i;
  ItListPage it;

  depth=1;
  ref = itPage->ref;
  for (i=m_vecPLsPgCache.size()-1; i>ref; i--)
    depth += m_vecPLsPgCache[i]->size();
  for (it=m_vecPLsPgCache[ref]->begin(); it!=itPage; it++)
    depth++;
  return depth;
}

// increase the statistic variables for cache hit
void Cache::inc(int idx, int n)
{
  if (!m_bWarmuping)
    m_vecStats[idx] += n;
}

// log the cache hit event
void Cache::LogHit(PEvent pEvent)
{
  inc(Hit);
  switch (pEvent->rw)
  {
  case READ:
    inc(RdHit);
    break;
  case WRITE:
    inc(WtHit);
    break;
  case ERASE:
    inc(ErHit); 
    break;
  default:
    assert(FALSE);
  }
}

// log the cache miss event
void Cache::LogMiss(PEvent pEvent)
{
    inc(Miss);
    switch (pEvent->rw)
    {
    case READ:
      inc(RdMiss);
      break;
    case WRITE:
      inc(WtMiss);
      break;
    case ERASE:
      inc(ErMiss); 
      break;
    default:
      assert(FALSE);
    }
}

// log the this event
void Cache::LogStats(PEvent pEvent)
{
  inc(Num);
  switch (pEvent->rw)
  {
  case READ:
    inc(Read);
    break;
  case WRITE:
    inc(Write);
    break;
  case ERASE:
    inc(Erase);
    break;
  default:
    assert(FALSE);
  }
}

/*
 * for debug.
 * check the integrity of the cache
 */
void Cache::check()
{
#ifdef _DEBUG
  int i;
  ItListPage it;
  PPage pPage;

  for (i=1; i<m_vecPLsPgCache.size(); i++)
  {
    for (it=m_vecPLsPgCache[i]->begin(); it!=m_vecPLsPgCache[i]->end(); it++)
    {
      pPage = &*it;
      if (m_hash.Search(pPage)==NULL)
      {
        cout<<"Page:"<<pPage->dev<<","<<pPage->inode<<","<<pPage->block<<" not found\n";
      }
    }
  }
#endif
}
