// Simulation.cpp: implementation of the Simulation class.
//
//////////////////////////////////////////////////////////////////////

#include "global.h"
#include "Simulation.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Simulation::Simulation()
{

}

Simulation::~Simulation()
{

}

//////////////////////////////////////////////////////////////////////
// public methods
//////////////////////////////////////////////////////////////////////

// get the next event from the trace
// return FALSE if no more event exists
BOOL Simulation::GetNextEvent(ifstream &ifInput, PEvent pEvent)
{
  char small_endian[sizeof(Event)];

  ifInput.read(small_endian,sizeof(Event));

#ifdef _MSC_VER // big_endian
  int i, n;
  char big_endian[sizeof(Event)];
  n = sizeof(long);
  for (i=0; i<sizeof(Event)/sizeof(long); i++)
  {
    big_endian[i*n+0] = small_endian[i*n+3];
    big_endian[i*n+1] = small_endian[i*n+2];
    big_endian[i*n+2] = small_endian[i*n+1];
    big_endian[i*n+3] = small_endian[i*n+0];
  }  
  memcpy(pEvent,big_endian,sizeof(Event));
#else // small_endian
  memcpy(pEvent,small_endian,sizeof(Event));
#endif

  return !ifInput.eof();
}

// read the config file and construct all Experiment objects
void Simulation::ReadConfig(char *strConfigName)
{
  Config cfg;
  PExperiment pExp;
  int i, iCltIdx, iSrvIdx, iAlgIdx;
  int iWarmup, iSeed;
  int iCltBuf, iSrvBuf, iAlg;
  char name[100];

  cfg.open(strConfigName); // open config file

  cfg.BeginParse();
  if (!cfg.VerifyKey("TraceFile"))
    Error("TraceFile expected");
  cfg.ParseStr(m_sTrace);

  cfg.BeginParse();
  if (!cfg.VerifyKey("Warmup"))
    Error("Warmup expected");
  cfg.ParseInt(iWarmup);
  m_iWarmup = iWarmup;

  cfg.BeginParse();
  if (!cfg.VerifyKey("Seed"))
    Error("Seed expected");
  cfg.ParseInt(iSeed);

  cfg.BeginParse();
  if (!cfg.VerifyKey("ClientBuffer"))
    Error("ClientBuffer expected");
  while (cfg.ParseInt(i) != EOD)
    m_vecCltBuf.push_back(i);

  cfg.BeginParse();
  if (!cfg.VerifyKey("ServerBuffer"))
    Error("ServerBuffer expected");
  while (cfg.ParseInt(i) != EOD)
    m_vecSrvBuf.push_back(i);

  cfg.BeginParse();
  if (!cfg.VerifyKey("ServerAlgorithm"))
    Error("ServerAlgorithm expected");
  while (stricmp(cfg.ParseStr(name),"EOD")!=0)
    m_vecAlg.push_back(iConvAlgName(name));

  cfg.close();

  // create all Experiments
  for (iCltIdx=0; iCltIdx < m_vecCltBuf.size(); iCltIdx++)
    for (iSrvIdx=0; iSrvIdx < m_vecSrvBuf.size(); iSrvIdx++)
      for (iAlgIdx=0; iAlgIdx < m_vecAlg.size(); iAlgIdx++)
      {
        iCltBuf = m_vecCltBuf[iCltIdx];
        iSrvBuf = m_vecSrvBuf[iSrvIdx];
        iAlg = m_vecAlg[iAlgIdx];
        pExp = new Experiment(iCltBuf, iWarmup, iSrvBuf, iAlg, iSeed);
        m_vecPExperiment.push_back(pExp);
      }
}

// feed the trace into all experiments
void Simulation::run()
{
  int i, num, n=0;
  Event event;

  ifstream ifTrace;

  // open the trace file
  ifTrace.open(m_sTrace, ios::binary);
  if (!ifTrace)
    Error("Can't open %s",name);

  // run all experiments
  num = m_vecPExperiment.size();
  while (GetNextEvent(ifTrace, &event))
  {
    n++;
    //if (n==8469)  // for debug
    //  n=n;
    // do all experiments for this event
    for (i=0; i<num; i++)
      m_vecPExperiment[i]->ProcessEvent(&event);

    //cout << n << ' ' << m_vecPExperiment[0]->pClient->m_vecStats[RdHit] << endl;
  }
  ifTrace.close();
}

// print the statistics data for all experiments
void Simulation::print(char *strOutputName)
{
  PExperiment pExp;
  ofstream ofOutput; // output file
  int i, num;

  // output results
  ofOutput.open(strOutputName, ios::app);
  if (!ofOutput)
    Error("Cannot open %s!\n",strOutputName);

  struct tm *newtime;
  time_t aclock;
  time( &aclock );                 /* Get time in seconds */
  newtime = localtime( &aclock );  /* Convert time to struct */
                                   /* tm form */
  /* Print local time as a string */
  ofOutput << asctime(newtime);

  ofOutput << "Warmup=" << m_iWarmup << endl;
  // output all results
  num = m_vecPExperiment.size();
  for (i=0; i<num; i++)
  {
    pExp = m_vecPExperiment[i];
    pExp->print(ofOutput);
  }
  ofOutput.close();

  // generate results for gnuplot
  //*
  int iCltBuf, iSrvBuf, iAlg;
  int iCltIdx, iSrvIdx, iAlgIdx;
  char sName[_MAX_PATH];
  double dRdMiss;

  iCltBuf = m_vecCltBuf.size();
  iSrvBuf = m_vecSrvBuf.size();
  iAlg = m_vecAlg.size();
  for (iAlgIdx=0; iAlgIdx < iAlg; iAlgIdx++)
    for (iCltIdx=0; iCltIdx < iCltBuf; iCltIdx++)
    {
      sprintf(sName, "clt_%s_%d.txt", GetAlgName(m_vecAlg[iAlgIdx]), m_vecCltBuf[iCltIdx]);
      ofOutput.open(sName);
      for (iSrvIdx=0; iSrvIdx < iSrvBuf; iSrvIdx++)
      {
        i = (iCltIdx*iSrvBuf+iSrvIdx)*iAlg+iAlgIdx;
        if (!( (m_vecPExperiment[i]->pClient->m_iSize==m_vecCltBuf[iCltIdx]) &&
                (m_vecPExperiment[i]->pServer->m_iSize==m_vecSrvBuf[iSrvIdx]) &&
                (m_vecPExperiment[i]->pServer->m_iAlg==m_vecAlg[iAlgIdx]) ))
          Error("Error in Simulation::Print()");
        VecInt &stats = m_vecPExperiment[i]->GetStats();
        dRdMiss = stats[RdMiss] * 1.0 / stats[Read];
        ofOutput << m_vecSrvBuf[iSrvIdx] << "\t" << dRdMiss << endl;
      }
      ofOutput.close();
    }
  for (iAlgIdx=0; iAlgIdx < iAlg; iAlgIdx++)
    for (iSrvIdx=0; iSrvIdx < iSrvBuf; iSrvIdx++)
    {
      sprintf(sName, "srv_%s_%d.txt", GetAlgName(m_vecAlg[iAlgIdx]), m_vecSrvBuf[iSrvIdx]);
      ofOutput.open(sName);
      for (iCltIdx=0; iCltIdx < iCltBuf; iCltIdx++)
      {
        i = (iCltIdx*iSrvBuf+iSrvIdx)*iAlg+iAlgIdx;
        if (!( (m_vecPExperiment[i]->pClient->m_iSize==m_vecCltBuf[iCltIdx]) &&
                (m_vecPExperiment[i]->pServer->m_iSize==m_vecSrvBuf[iSrvIdx]) &&
                (m_vecPExperiment[i]->pServer->m_iAlg==m_vecAlg[iAlgIdx]) ))
          Error("Error in Simulation::Print()");
        VecInt &stats = m_vecPExperiment[i]->GetStats();
        dRdMiss = stats[RdMiss] * 1.0 / stats[Read];
        ofOutput << m_vecCltBuf[iCltIdx] << "\t" << dRdMiss << endl;
      }
      ofOutput.close();
    }
  //*/

  // delete all experiment objects
  for (i=0; i<num; i++)
  {
    pExp = m_vecPExperiment[i];
    delete pExp;
  }
}

//////////////////////////////////////////////////////////////////////
// protected methods
//////////////////////////////////////////////////////////////////////

// convert the server replacement algorithm from string to integer
int Simulation::iConvAlgName(char *name)
{
  if (stricmp(name,"LRU")==0)
    return LRU;
  if (stricmp(name,"LFU")==0)
    return LFU;
  if (stricmp(name,"RAND")==0)
    return RAND;
  if (stricmp(name,"LruDep")==0)
    return LRUDEP;
  if (stricmp(name,"LfuDep")==0)
    return LFUDEP;
  Error("Wrong name:%s",name);
  return -1;
}

char* Simulation::GetAlgName(int iAlg)
{
  switch (iAlg)
  {
  case LFU:
    return "lfu";
  case LRU:
    return "lru";
  case RAND:
    return "ran";
  default:
    assert(FALSE);
  }
  return NULL;
}
