#include "defines.h"
#include "file_io.h"
#include <assert.h>

/* one file pointer for each client's trace file -- thus limits the number */
/* of clients we can have to max number of files a process is allowed to */
/* have open concurrently */

FILE *tr_files[MAX_CLIENTS+1];


/* given the name of a file which contains the names of the trace files */
/* (one filename per line with 1st name corresponding to client 1's trace */
/* file, 2nd name to client 2's file, etc., read the file and place each */
/* name into an array. Return the number of names read. */

/* NOTE: first name is placed in filenames[1] not filenames[0] */
char filenames[MAX_CLIENTS+1][MAX_FILENAME_SIZE+1];  

/* set to X when the end of client X's trace file is reached */
int end_of_any_file = FALSE;

int read_filenames(filename)
char filename[];
{
  FILE *fp;
  int count = 0, nd;
  
  fp = fopen(filename, "r");
  if (fp == NULL) {
    printf("ERROR: cannot open file %s\n",filename);
    exit(1);
  }
  
  /* read each filename until end of file or MAX_CLIENTS words are read */
  do {
       /* just hope that all words are <= MAX_FILENAME_SIZE or strange things
    happen */
    nd = fscanf(fp,"%s",filenames[count+1]);
    if (nd == EOF) return(count);

    if (nd != 1) {
      printf("ERROR: error reading name from %s\n",filename);
      exit(1);
    }
    count++;
  } while (count < MAX_CLIENTS);

  return(count);
}

/* open trace file for a client  */
void open_trace_file(client)
{
  tr_files[client] = fopen(filenames[client], "r");
  if (tr_files[client] == NULL ) {
    printf("ERROR: cannot open file %s\n",filenames[client]);
    exit(1);
  }
}


/*

  read the next reference in the trace file for a client (client i
uses file i) and return a structure describing it.  To incorporate the
fact that some references are to shared files while others are to
non-shared files, the trace contains a field called 'flag' which is 0
if the reference is to a non-shared file, and -1 if it is to a shared
file (a flag=-2 and wr=-1 means this is not a reference, but an
'unlink' event; a flag > 0 means that this is a getblk or vnsremove
event we are reading from a server reference stream that the
simulation had written on some previous pass -- if wr=-1, its a
vnsremove, if wr = 0 or 1 or 2, its a getblk).  If we see a flag value
of 0, we set it to be equal to the client id so that it will never
match any references with the same (dev, inode, blkno) triple from
another client's trace.  (i.e., we treat the similar references as
being to different files).  Similarly, for vnsremove events, we set
flag=client id (if not already), so that it only matches blocks for
this client (THE assumption is that we never remove shared files).
The simulator can still distinguish a vnsremove from a getblk by
looking at the wr value (0,1 or 2 means its a getblk; -1 means its a
vnsremove).

 */ 

struct trace_node read_trace(client)
int client; 
{ 
  struct trace_node input; 
  int check;

  check = fread(&input, sizeof(struct trace_node), 1, tr_files[client] );

  if (check != 1) {
    if (feof(tr_files[client])) {
      end_of_any_file = client;
      return(input);   /* note: this has not been filled and should not be used anywhere */
    }
    else {
      printf("ERROR: error reading trace file %s\n",filenames[client]);
      exit(1);
    }
  }

  assert(input.flag >= 0 || 
   input.flag == -1 || 
   (input.flag == -2 && input.wr == -1) );


  /* set the flag field = client if this is supposed to be a getblk reference */
  /* to a non-shared file or if it is a vnsremove event. */
  if (input.flag == 0) input.flag = client;
  if (input.flag == -2) input.flag = client;

  return(input);
}
